(ns com.aurellem.run.bootstrap-0
  (:use (com.aurellem.gb gb-driver vbm characters))
  (:use (com.aurellem.run title save-corruption))
  (:use (com.aurellem.exp item-bridge))
  (:import [com.aurellem.gb.gb_driver SaveState]))

(defn-memo boot-root []
  [ [] (root)])
  
(defn-memo to-rival-name
  ([] (to-rival-name (boot-root)))
  ([script]
     (-> script
         title
         oak
         name-entry-rlm
         scroll-text
         scroll-text
         scroll-text
         scroll-text
         scroll-text
         )))

(defn-memo name-rival-bootstrap
  ([] (name-rival-bootstrap (to-rival-name)))
  ([script]
     (->> script
         (advance [] [:a])
         (advance [] [:r] DE)
         (play-moves
          [[]
           [:r] [] [:r] [] [:r] [] [:r] []
           [:r] [] [:r] [] [:r] [] [:d] []
           [:d] [:a]                         ;; space
           [:l] [] [:d] [:a]                 ;; [PK]
           [:u] [] [:u] [] [:u] [] [:l] [:a] ;; G
           [:d] [] [:d] [] [:d] [] [:r] [:a] ;; [PK]
           [:u] [] [:u] [] [:u] [] [:l] [:a] ;; G
           [:d] [] [:d] [] [:d] [] [:r] [:a] ;; [PK]

           [:d] [] [:r] [:a]                 ;; finish
           ]))))

(defn walk
  "Move the character along the given directions."
  [directions script]
  (reduce (fn [script direction]
            (move direction script))
          script directions))

(def ↑ [:u])
(def ↓ [:d])
(def ← [:l])
(def → [:r])

(defn-memo leave-house
  ([] (leave-house (name-rival-bootstrap)))
  ([script]
     (->> script
          finish-title
          start-walking
          walk-to-stairs
          walk-to-door
          (walk [↓ ↓]))))

(defn-memo to-pallet-town-edge
  ([] (to-pallet-town-edge (leave-house)))
  ([script]
     (->> script
          start-walking
          (walk [→ → → → →
                 ↑ ↑ ↑ ↑ ↑ ↑]))))

(defn end-text [script]
  (->> script
       (scroll-text)
       (play-moves [[] [:a]])))

(defn-memo start-pikachu-battle
  ([] (start-pikachu-battle
       (to-pallet-town-edge)))
  ([script]
     (->> script
          (advance [:b] [:b :a] DE)
          (scroll-text)
          (play-moves [[:b]])
          (scroll-text)
          (end-text) ;; battle begins
          (scroll-text))))

(defn-memo capture-pikachu
  ([] (capture-pikachu (start-pikachu-battle)))
  ([script]
     (->> script
          (scroll-text 2)
          (end-text))))

(defn-memo go-to-lab
  ([] (go-to-lab (capture-pikachu)))
  ([script]
     (->> script
          (scroll-text 5)
          (end-text)
          (scroll-text)
          (end-text)
          (scroll-text 8)
          (end-text)
          (scroll-text)
          (end-text))))

(defn-memo obtain-pikachu
  ([] (obtain-pikachu (go-to-lab)))
  ([script]
     (->> script
          (scroll-text)
          (play-moves
           (concat 
            (repeat 51 [])
            [[:a] []]))
          (walk [↓ ↓ → → ↑])
          (play-moves
           (concat [[] [:a]]
                   (repeat 100 [])))
          (scroll-text 9)
          (end-text)
          (scroll-text 7)

          (play-moves
           (concat
            (repeat 42 [])
                   [[:b] [:b] [:b] [:b]])))))

(defn-memo begin-battle-with-rival
  ([] (begin-battle-with-rival
        (obtain-pikachu)))
  ([script]
     (->> script
          (walk [↓ ↓ ↓ ↓])
          (scroll-text 3)
          (end-text)
          (scroll-text))))

(defn search-string
  [array string]
  (let [codes
        (str->character-codes string)
        codes-length (count codes)
        mem (vec array)
        mem-length (count mem)]
    (loop [idx 0]
      (if (< (- mem-length idx) codes-length)
        nil
        (if (= (subvec mem idx (+ idx codes-length))
               codes)
          idx
          (recur (inc idx)))))))
        
(defn critical-hit
  "Put the cursor over the desired attack. This program will
   determine the appropriate amount of blank frames to
   insert before pressing [:a] to ensure that the attack is
   a critical hit."
  [script]
  (loop [blanks 6]
    (let [new-script
          (->> script
               (play-moves
                (concat (repeat blanks [])
                        [[:a][]])))]
      (if (let [future-state
                (run-moves (second new-script)
                           (repeat 400 []))

                result (search-string (memory future-state)
                                      "Critical")]
            (if result
              (println "critical hit with" blanks "blank frames"))
            result) 
        new-script
        (recur (inc blanks))))))  

(defn-memo battle-with-rival
  ([] (battle-with-rival
        (begin-battle-with-rival)))
  ([script]
     (->> script
          (play-moves (repeat 381 []))
          (play-moves [[:a]])
          (critical-hit)
          (play-moves (repeat 100 []))
          (scroll-text)
          (play-moves
           (concat (repeat 275 []) [[:a]]))
          (critical-hit)
          (play-moves (repeat 100 []))
          (scroll-text)
          (play-moves
           (concat (repeat 270 []) [[:a]]))
          (play-moves [[][][][][][][][][:a]]))))

(defn-memo finish-rival-text
  ([] (finish-rival-text
       (battle-with-rival)))
  ([script]
     (->> script
          (scroll-text 2)
          (end-text)
          (scroll-text 9)
          (end-text))))

(defn do-nothing [n script]
  (->> script
       (play-moves
        (repeat n []))))

(defn-memo pikachu-comes-out
  ([] (pikachu-comes-out
       (finish-rival-text)))
  ([script]
     (->> script
          (do-nothing 177)
          (end-text)
          (scroll-text 7)
          (end-text))))

(defn-memo leave-oaks-lab
  ([] (leave-oaks-lab
       (pikachu-comes-out)))
  ([script]
     (->> script 
          (walk [← ← ↓ ↓ ↓ ↓ ↓ ↓]))))

(defn-memo  oaks-lab->pallet-town-edge
  ([] (oaks-lab->pallet-town-edge
       (leave-oaks-lab)))
  ([script]
     (->> script
          (walk [← ← ← ←
                 ↑ ↑ ↑ ↑
                 ↑ ↑ ↑ ↑ ↑ ↑
                 → ↑]))))
                 

(defn move-thru-grass
  [direction script]
  (loop [blanks 0]
    (let [new-script
          (->> script
               (play-moves (repeat blanks []))
               (move direction))

          future-state
          (run-moves (second new-script)
                     (repeat 600 []))
          
          result (search-string (memory future-state)
                                "Wild")]
      (if (nil? result)
        new-script
        (recur (inc blanks))))))

(defn walk-thru-grass
  [directions script]
  (reduce (fn [script direction]
            (move-thru-grass direction script))
          script directions))

(defn-memo pallet-edge->viridian-mart
  ([] (pallet-edge->viridian-mart
       (oaks-lab->pallet-town-edge)))
  ([script]
     (->> script
          ;; leave straight grass
          (walk-thru-grass
           [↑ ↑ ↑ ↑ ↑])
          
          (walk [↑ ↑ ↑ ↑])
          
          (walk-thru-grass
           [← ← ↑])
          (walk [↑ ↑ ↑ ↑ → → → ])

          (walk-thru-grass
           [→ ↑ ↑ ←])

          (walk
           [← ←
            ↑ ↑ ↑ ↑ ↑ ↑ ↑ ↑
            → → → → ])

          ;; this part it dependent on that
          ;; stupid NPC in the grass patch
          (walk-thru-grass
           [→ →
            ↑ ↑ ↑ ↑ ↑ ↑ ↑ ↑ ])
           
          (walk
           [↑ ↑
            ← ← ← ←
            ↑ ↑ ↑ ↑ ↑ ↑ ↑ ↑ ↑ ↑
            ← ←
            ↑ ↑ ↑ ↑
            → → → → → → → → → →
            ↑ ↑ ↑ ↑ ↑ ↑ ↑])))) 
            
(defn-memo get-oaks-parcel
  ([] (get-oaks-parcel
       (pallet-edge->viridian-mart)))
  ([script]
     (->> script
          (end-text)
          (scroll-text 3)
          (do-nothing 197)
          (play-moves [[:a] []])
          (walk [↓ ↓ → ↓]))))

(defn-memo viridian-store->oaks-lab
  ([] (viridian-store->oaks-lab
       (get-oaks-parcel)))
  ([script]
     (->> script
          (walk [↓ ↓ ↓ ↓ ↓ ↓ ↓ ↓ ↓ ↓
                 ← ← ← ← ← ← ← ← ← ← 
                 ↓ ↓ ↓ ↓ ↓ ↓ ↓ ↓
                 ← ←
                 ↓ ↓ ↓ ↓ ↓ ↓ ↓ ↓
                 ↓ ↓ ↓ ↓ ↓ ↓ ↓
                 → → → → → → → →
                 ↓ ↓ ↓ ↓
                 ← ← ← ← ←
                 ↓ ↓ ↓ ↓])

          (walk-thru-grass
           [↓ ↓ ↓ ↓ ↓ ↓ ↓])

          (walk [↓ ↓ ← ↓ ↓ ↓ ←
                 ↓ ↓ ↓ ↓ ↓
                 → → → ↑]))))
                 
(defn-memo viridian-store->oaks-lab-like-a-boss
  ([] (viridian-store->oaks-lab-like-a-boss
       (get-oaks-parcel)))
  ([script]
     (->> script
          (walk [↓ ↓ ↓ ↓ ↓ ↓ ↓ ↓ ↓ ↓
                 ← ← ← ← ← ← ← ← ← ← 
                 ↓ ↓ ↓ ↓ ↓ ↓ ↓ ↓ ↓])

          (walk-thru-grass
           [↓ ↓ ↓ ↓ ↓])

          (walk
           [↓ ↓ ← ↓
            ↓ ↓ ↓ ↓ ↓ ↓ ↓ ↓
            → →])

          (walk-thru-grass
           [→ ↓ ↓ ↓])

          (walk [↓ ← ← ↓ ↓ ↓ ↓ ↓ ↓])
          
          (walk-thru-grass
           [↓ ↓ ↓ ↓ ↓ ↓ ↓])
          
          (walk [↓ ↓ ← ↓ ↓ ↓ ←
                 ↓ ↓ ↓ ↓ ↓
                 → → → ↑]))))
          